/*
  dxref - a tiny cross referencer using DWARF2-XML
  Copyright (C) 2003, 2004, by gondow@jaist.ac.jp, All rights reserved.
  $Id: dxref.c,v 1.13 2004/01/20 14:22:10 gondow Exp gondow $
 */
/* ======================================================================== */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>

#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/tree.h>
#include <libxml/debugXML.h>

#define BUFLEN            8192

#define TARGET_MAIN       "target=\"main\""
#define TARGET_SRC        "target=\"src\""
#define COMM_BEGIN        "<font color=\"red\">/* "
#define COMM_END          " */</font>"

#define DEFAULT_DIR       "./HTML/"
#define WORKING_DIR       "./HTML/dxref.files/"
#if 0
#define XREF_MAIN         "dxref_main.html"
#endif
#define XREF_INFO         "dxref_info.html"
#define XREF_MACINFO      "dxref_macinfo.html"
#define XREF_INDEX        "dxref_index.html"
#define XREF_STRUCTS1     "dxref_structs1.html"
#define XREF_STRUCTS2     "dxref_structs2.html"
#define XREF_UNIONS1      "dxref_unions1.html"
#define XREF_UNIONS2      "dxref_unions2.html"
#define XREF_TYPEDEFS1    "dxref_typedefs1.html"
#define XREF_TYPEDEFS2    "dxref_typedefs2.html"
#define XREF_ENUM_TAGS1   "dxref_enum_tags1.html"
#define XREF_ENUM_TAGS2   "dxref_enum_tags2.html"
#define XREF_ENUM_CONSTS1 "dxref_enum_consts1.html"
#define XREF_ENUM_CONSTS2 "dxref_enum_consts2.html"
#define XREF_LABELS1      "dxref_labels1.html"
#define XREF_LABELS2      "dxref_labels2.html"
#define XREF_FUNCTIONS1   "dxref_functions1.html"
#define XREF_FUNCTIONS2   "dxref_functions2.html"
#define XREF_VARIABLES1   "dxref_variables1.html"
#define XREF_VARIABLES2   "dxref_variables2.html"
#define XREF_MACROS1      "dxref_macros1.html"
#define XREF_MACROS2      "dxref_macros2.html"
#define XREF_MACROS3      "dxref_macros3.html"
#define XREF_FILES1       "dxref_files1.html"
#define XREF_FILES2       "dxref_files2.html"

/* ====================================================================== */
static struct {
    int src2html;
    int  num_files;
    char **files;
    int verbose;
} args = {
    0,   /* src2html */
    0,   /* num_files */
    NULL,/* files */
    0    /* verbose */
};

struct comp_unit {
    char *info_offset;
    unsigned long info_offset2;
    char *macinfo_offset;
    char *line_offset;
    int n_file;
    struct file *file; /* array */
    int n_line;
    struct line *line; /* array */
};

struct file {
    char *filename;
    char *pathname;
};

struct line {
    unsigned long addr;
    int line;
    struct file *file;
};

struct abbrev_entry {
    int    number;
    char   *upper_bound_form;
};
/* ====================================================================== */
static xmlDocPtr  xml_doc  = NULL;
static xmlNodePtr xml_root = NULL;
static int    indent_depth = 0;
static int n_comp_unit;
static struct comp_unit *comp_unit;
static struct abbrev_entry *abbrev_table;
static int eval_xpath_debug = 0; 
/* ====================================================================== */
static void *emalloc (size_t size);
static void print_usage (void);
static void process_args (int argc, char *argv []);
static void init (void);
static void final (void);
static int count_children (xmlNodePtr node, char *tag);
static xmlNodePtr search_sibling (xmlNodePtr node, char *tag);
static char **create_dir_table (xmlNodePtr node);
static void delete_dir_table (char **dir_table);
static struct file *create_file_table (xmlNodePtr node, char **dir_table,
				       int *n_file);
static struct line *create_line_table (xmlNodePtr node,
				       struct file *file_table, int *n_line);
static void create_comp_unit_table (void);
static void create_abbrev_table (void);
static void do_src2html (void);

static char *href_info (char *id, char *title);
static char *href_info2 (char *id, char *title, char *title2);
static char *href_macinfo (char *id, char *title);
static char *href_macinfo2 (char *id, char *title, char *title2);
static char *href_src1 (char *pathname, char *line, char *title);
static char *href_src2 (char *pathname, char *line, char *title);
static char *a_name    (char *id);

static xmlNodeSetPtr eval_xpath   (char *xpath, xmlNodePtr from,
				   void **free_p, int *num);
static inline char *get_child     (char *elem, char *name, char *attr,
				   xmlNodePtr from);
static inline char *get_tag       (char *name, char *attr, xmlNodePtr from);
static inline char *get_attr      (char *name, char *attr, xmlNodePtr from);
static inline char *get_attr_value     (char *xpath, xmlNodePtr from);
static inline char *get_attr_value_ref (char *name, xmlNodePtr from);
static inline char *get_attr_desc      (char *xpath, xmlNodePtr from);

static char *get_type             (xmlNodePtr node);
static char *get_params           (xmlNodePtr node);
static char *get_subscript        (xmlNodePtr node);
static int  get_comp_unit         (xmlNodePtr node);
static char *get_filename_by_char (int n_comp, char *file_num);
static char *get_filename_by_node (xmlNodePtr node, char *file_num);
static char *get_filename_by_int  (int n_comp, int file_num);
static char *get_pathname_by_char (int n_comp, char *file_num);
static char *get_pathname_by_node (xmlNodePtr node, char *file_num);
static char *get_pathname_by_int  (int n_comp, int file_num);
static char *get_line             (char *line);
static char *get_name             (char *name);

static char *type_to_string       (xmlNodePtr node);
static char *base_type_to_string  (xmlNodePtr node);
static char *array_type_to_string (xmlNodePtr node);
static char *ptr_type_to_string   (xmlNodePtr node);
static char *func_type_to_string  (xmlNodePtr node);
static char *volatile_to_string   (xmlNodePtr node);
static char *const_to_string      (xmlNodePtr node);
static char *struct_union_to_string (xmlNodePtr node, char *which);
static char *struct_to_string     (xmlNodePtr node);
static char *union_to_string      (xmlNodePtr node);
static char *typedef_to_string    (xmlNodePtr node);
static char *enum_to_string       (xmlNodePtr node);
static char *enum_constants_to_string (xmlNodePtr node);
static char *members_to_string    (xmlNodePtr node);

static void visit_all_children    (xmlNodePtr node, FILE *fp);
static void output_indent         (FILE *fp);
static void output_debug_info_top (void);
static void output_debug_info     (xmlNodePtr node, FILE *fp);
static void output_compile_unit   (xmlNodePtr node, FILE *fp);
static void output_subprogram     (xmlNodePtr node, FILE *fp);
static void output_lexical_block  (xmlNodePtr node, FILE *fp);
static void output_variable       (xmlNodePtr node, FILE *fp);
static void output_label          (xmlNodePtr node, FILE *fp);
static void output_typedef        (xmlNodePtr node, FILE *fp);
static void output_enum           (xmlNodePtr node, FILE *fp);
static void output_enum_constants (xmlNodePtr node, FILE *fp,
				   char *pathname, char *line);
static void output_struct_union   (xmlNodePtr node, FILE *fp, char *which);
static void output_struct         (xmlNodePtr node, FILE *fp);
static void output_union          (xmlNodePtr node, FILE *fp);
static void output_members        (xmlNodePtr node, FILE *fp);

static void output_all_structs    (void);
static void output_all_unions     (void);

static void output_all_typedefs   (void);
static void output_all_enums      (void);
static void output_all_labels     (void);
static void output_all_functions  (void);
static void output_all_variables  (void);
static void output_all_files      (void);

static char *get_macinfo_id (void);
static void output_all_macros     (void);
/* ====================================================================== */
static void *
emalloc (size_t size)
{ 
    void *p = malloc (size);
    if (p == NULL) perror ("malloc");
    return p;
}

static void
print_usage (void)
{
    fprintf (stderr, 
	     "usage  : dxref [options] file.xml\n"
	     "options:\n"
	     "    --src2html           translate sources to htmls\n"
	     "    --verbose            tells what is being doing\n"
	     "    --help               prints this usage\n"
	     );
}

static void
process_args (int argc, char *argv [])
{
    int    i;
    char   *s;

    for (i = 1; i < argc; i++) {
	if (argv [i][0] != '-') { /* filenames */
	    args.files = &argv [i];
	    args.num_files = argc - i;
	    break;
	} else if (!strcmp (argv [i], "--src2html")) { 
	    args.src2html = 1;
	} else if (!strcmp (argv [i], "--verbose")) { 
	    args.verbose  = 1;
	} else if (!strcmp (argv [i], "--help")) {
	    print_usage ();
	    exit (1);
	} else {
	    /* unknown option */
	    fprintf (stderr, "unknown option: %s\n", argv [i]);
	    print_usage ();
	    exit (1);
	}
    }

    if (args.num_files <= 0) {
	print_usage ();
	exit (1);
    }
}

static void
init (void)
{
    char *xml;

    assert (args.num_files >= 1);
    xml = args.files [0];    /* currently, only one file is processed */
    xmlDoValidityCheckingDefaultValue = 1;
    xmlSubstituteEntitiesDefault (1); /* to substitute entity refs. (&lt;) */

    if (!strcmp (".gz", xml + strlen (xml) -3)) { /* gzipped file */
	FILE *fp;
	char buf[BUFLEN], *gunzip_file;
	int  ret;

	gunzip_file = tmpnam (NULL);
	snprintf (buf, sizeof (buf), "gunzip -c %s > %s", xml, gunzip_file);
	ret = system (buf);
	if (ret == -1) {
	    perror ("system"); exit (1);
	}
	fp = fopen (gunzip_file, "r");
	xml_doc  = xmlParseFile(gunzip_file);
	unlink (gunzip_file);
    } else { /* not gzipped file */
	xml_doc  = xmlParseFile(xml);
    }
    xml_root = xmlDocGetRootElement (xml_doc);
    xmlXPathInit();
    mkdir (DEFAULT_DIR, 0755);
    mkdir (WORKING_DIR, 0755);
    create_comp_unit_table ();
    create_abbrev_table ();

    /* must be after create_comp_unit_table (); */
    if (args.src2html) do_src2html ();
}

static void
final (void)
{
    system ("dxref.csh");
}

static int
count_children (xmlNodePtr node, char *tag)
{
    int n = 0;
    xmlNodePtr child;

    for (child = node->children; child != NULL; child = child->next) {
	if (child->type == XML_ELEMENT_NODE && !strcmp (child->name, tag))
	    n++;
    }
    return n;
}

static xmlNodePtr
search_sibling (xmlNodePtr node, char *tag)
{
    while (node != NULL) {
	if (node->type == XML_ELEMENT_NODE && !strcmp (node->name, tag))
	    return node;
	node = node->next;
    }
    return NULL; /* not found */
}

static char **
create_dir_table (xmlNodePtr node)
{
    int i, n_dir;
    xmlNodePtr child;
    char **dir_table;

    assert (node != NULL);
    assert (node->type == XML_ELEMENT_NODE);
    assert (!strcmp (node->name, "include_directories"));

    n_dir = count_children (node, "directory");
    /* +1 for 1-origin, +1 for NULL-termination */
    dir_table = emalloc (sizeof (char *) * (n_dir + 2));
    dir_table [0] = "!!!not used entry";
    i = 1;
    for (child = node->children; child != NULL; child = child->next) {
	if (child->type == XML_ELEMENT_NODE
	    && !strcmp (child->name, "directory"))
	    dir_table [i++] = xmlNodeGetContent (child);
    }
    dir_table [i] = NULL;

    return dir_table;
}

static void
delete_dir_table (char **dir_table)
{
    int i;

    if (dir_table == NULL)
	return;

    for (i = 1; dir_table [i] != NULL; i++) {
	xmlFree (dir_table [i]);
    }
    free (dir_table);
}

static struct file *
create_file_table (xmlNodePtr node, char **dir_table, int *n_file)
{
    int i;
    struct file *file_table;
    xmlNodePtr child;
    char buf [BUFLEN];

    *n_file = count_children (node, "file");
    /* +1 for 1-origin */
    file_table = emalloc (sizeof (struct file) * *n_file + 1);

    file_table [0].filename = "!!!unused entry";
    file_table [0].pathname = "!!!unused entry";
    i = 1;
    for (child = node->children; child != NULL; child = child->next) {
	if (child->type == XML_ELEMENT_NODE && !strcmp (child->name, "file")) {
	    char *dir_index = xmlGetProp (child, "directory");
	    int  d = atoi (dir_index);
	    xmlFree (dir_index);

	    file_table [i].filename = xmlNodeGetContent (child);

	    if (d == 0) {
		file_table [i].pathname = file_table [i].filename;
	    } else {
		snprintf (buf, sizeof (buf), "%s/%s",
			  dir_table [d], file_table [i].filename);
		file_table [i].pathname = strdup (buf);
	    }
	    i++;
	}
    }
    assert (i == *n_file + 1);
    return file_table;
}

static struct line *
create_line_table (xmlNodePtr node, struct file *file_table, int *n_line)
{
    int i;
    struct line *line_table;
    xmlNodePtr child;

    *n_line = count_children (node, "line");
    line_table = emalloc (sizeof (struct line) * *n_line);

    i = 0;
    for (child = node->children; child != NULL; child = child->next) {
	if (child->type == XML_ELEMENT_NODE && !strcmp (child->name, "line")) {
	    char *addr = xmlGetProp (child, "address");
	    char *line = xmlGetProp (child, "line");
	    char *file = xmlGetProp (child, "file");

	    line_table [i].addr = atoll (addr);
	    line_table [i].line = atoi (line);
	    line_table [i].file = &file_table [atoi (file)];

	    xmlFree (addr);
	    xmlFree (line);
	    xmlFree (file);
	    i++;
	}
    }
    assert (i == *n_line);
    return line_table;
}

static void
create_comp_unit_table (void)
{
    int  i;
    char *xpath = "//section[@name='.debug_info']/tag[@name='DW_TAG_compile_unit']";
    void *free_p;

    xmlNodeSetPtr result = eval_xpath (xpath, xml_root, &free_p,
				       &n_comp_unit);
    assert (n_comp_unit >= 1);
    comp_unit = emalloc (sizeof (struct comp_unit) * n_comp_unit);

    /* store debug line info. etc for each compilation_unit */
    for (i = 0; i < n_comp_unit; i++) {
	xmlNodePtr node = result->nodeTab [i];
	char **dir_table;
	xmlNodePtr line_node, line_node2;
	char buf [BUFLEN];
	xmlAttrPtr attr;
	int ret;

	char *info_offset = xmlGetProp (node, "offset");
	char *macro_info  = get_attr_value ("DW_AT_macro_info", node);
	char *stmt_list   = get_attr_value ("DW_AT_stmt_list", node);

	assert (info_offset != NULL);
	assert (macro_info  != NULL);
	assert (stmt_list   != NULL);

	comp_unit [i].info_offset = info_offset;
	ret = sscanf (info_offset, "info:%ld", &comp_unit [i].info_offset2);
	if (ret != 1)
	    fprintf (stderr, "info_offset = %s\n", info_offset);
	assert (ret == 1);
	snprintf (buf, sizeof (buf), "macinfo:%s", macro_info);
	comp_unit [i].macinfo_offset = strdup (buf);
	snprintf (buf, sizeof (buf), "line:%s",    stmt_list);
	comp_unit [i].line_offset = strdup (buf);

	attr = xmlGetID (xml_doc, comp_unit [i].line_offset);
	assert (attr != NULL);
	line_node = attr->parent;

	line_node2 = search_sibling (line_node, "include_directories");
	if (line_node2 == NULL) { /* no directory info. */
	    dir_table = NULL;
	} else {
	    line_node = line_node2;
	    dir_table = create_dir_table (line_node);
	}
	line_node = search_sibling (line_node, "file_names");
	assert (node != NULL);
	comp_unit [i].file = create_file_table (line_node, dir_table,
						 &comp_unit [i].n_file);
	delete_dir_table (dir_table);
#if 1
	comp_unit [i].line = NULL;
#else
	/* currently, dxref does not use `comp_unit [i].line', so
	  I commented out the following. */
	line_node = search_sibling (line_node, "lines");
	assert (node != NULL);
	comp_unit [i].line = create_line_table (line_node, comp_unit [i].file,
						 &comp_unit [i].n_line);
#endif
	xmlFree (macro_info);
	xmlFree (stmt_list);
    }
}

static void
create_abbrev_table (void)
{
    int  i, n;
    char *xpath = "//section[@name='.debug_abbrev']/tag_abbrev"
	          "[@name='DW_TAG_subrange_type']/"
    	          "attribute_abbrev[@name='DW_AT_upper_bound']";
    void *free_p;
    xmlNodeSetPtr result = eval_xpath (xpath, xml_root, &free_p, &n);

    abbrev_table = emalloc (sizeof (struct abbrev_entry) * (n + 1));
    for (i = 0; i < n; i++) {
	xmlNodePtr node = result->nodeTab [i];
	char *number = xmlGetProp (node->parent, "number");
	abbrev_table [i].number = atoi (number);
	xmlFree (number);
	abbrev_table [i].upper_bound_form = xmlGetProp (node, "type");
    }
    abbrev_table [n].number = -1; /* sentinel */
    abbrev_table [n].upper_bound_form = NULL;
    free (free_p);
}

static void
do_src2html (void)
{
    int i, j, ret;
    char buf [BUFLEN];

    if (args.verbose) printf ("translating sources to htmls\n");
    for (i = 0; i < n_comp_unit; i++) {
	for (j = 1; j <= comp_unit [i].n_file; j++) {
	    snprintf (buf, sizeof (buf), "PATH=.:$PATH; src2html %s",
		      comp_unit [i].file [j].pathname);
	    ret = system (buf);
	    if (ret == -1) { perror ("system"); exit (1); }
	}
    }
}

static char *
href_info (char *id, char *title)
{
    static char buf [BUFLEN];
    snprintf (buf, sizeof (buf), "<a href=\"" XREF_INFO "#%s\" "
	      TARGET_MAIN ">%s</a>", id, title);
    return buf;
}

static char *
href_info2 (char *id, char *title, char *title2)
{
    static char buf [BUFLEN];
    snprintf (buf, sizeof (buf), "<a href=\"" XREF_INFO "#%s\" "
	      TARGET_MAIN ">%s %s</a>", id, title, title2);
    return buf;
}

static char *
href_macinfo (char *id, char *title)
{
    static char buf [BUFLEN];
    snprintf (buf, sizeof (buf), "<a href=\"" XREF_MACINFO "#%s\" "
	      TARGET_MAIN ">%s</a>", id, title);
    return buf;
}

static char *
href_macinfo2 (char *id, char *title, char *title2)
{
    static char buf [BUFLEN];
    snprintf (buf, sizeof (buf), "<a href=\"" XREF_MACINFO "#%s\" "
	      TARGET_MAIN ">%s %s</a>", id, title, title2);
    return buf;
}

static char *
href_src1 (char *pathname, char *line, char *title)
{
    static char buf [BUFLEN];
    snprintf (buf, sizeof (buf), "<a href=\"./%s.html#%s\" "
	      TARGET_SRC ">%s</a>", pathname, line, title);
    return buf;
}

static char *
href_src2 (char *pathname, char *line, char *title)
{
    static char buf [BUFLEN];
    snprintf (buf, sizeof (buf),
	      "<a href=\"./%s.html#%s\" " TARGET_SRC ">%s:%s</a>",
	      pathname, line, title, line);
    return buf;
}

static char *
a_name    (char *id)
{
    static char buf [BUFLEN];
    snprintf (buf, sizeof (buf), "<a name=\"%s\"/>", id);
    return buf;
}

static xmlNodeSetPtr 
eval_xpath (char *xpath, xmlNodePtr from, void **free_p, int *num)
{
    xmlXPathObjectPtr  result;
    xmlXPathContextPtr context;

    context = xmlXPathNewContext(xml_doc);
    assert (context != NULL);
    context->node = from;
    result = xmlXPathEval (xpath, context);
    xmlXPathFreeContext (context);
#if 0
    if (args.verbose)
	xmlXPathDebugDumpObject (stdout, result, 0);
#endif
    if (num != NULL)
	*num = xmlXPathNodeSetGetLength (result->nodesetval);
    *free_p = result;
    return result->nodesetval;
}

static inline char *
get_child (char *elem, char *name, char *attr, xmlNodePtr from)
{
    int i;
    xmlNodePtr node;
    char *attr_name;

    assert (from->children != NULL);
    for (node = from->children; node != NULL; node = node->next) {
	if (node->type != XML_ELEMENT_NODE) continue;
	if (strcmp (node->name, elem))      continue;
	attr_name = xmlGetProp (node, "name");
	if (strcmp (attr_name, name)) {
	    xmlFree (attr_name);
	    continue;
	}
	xmlFree (attr_name);
	if (*attr == '@') {
	    attr++;
	    return xmlGetProp (node, attr);
	} else {
	    xmlNodePtr child;
	    for (child = node->children; child != NULL; child = child->next) {
		if (child->type != XML_ELEMENT_NODE) continue;
		if (strcmp (child->name, attr))      continue;
		return xmlNodeGetContent (child);
	    }
	    return NULL;
	}
    }
    return NULL;
}

static inline char *
get_tag (char *name, char *attr, xmlNodePtr from)
{
    return get_child ("tag", name, attr, from);
}

static inline char *
get_attr (char *name, char *attr, xmlNodePtr from)
{
    return get_child ("attribute", name, attr, from);
}

static inline char *
get_attr_value (char *name, xmlNodePtr from)
{
    return get_attr (name, "@value", from);
}

static inline char *
get_attr_value_ref (char *name, xmlNodePtr from)
{
    return get_attr (name, "@value_ref", from);
}

static inline char *
get_attr_desc (char *name, xmlNodePtr from)
{
    return get_attr (name, "description", from);
}

static char *
get_type (xmlNodePtr node)
{
    char *type;
    char *type_id;

    type_id = get_attr_value_ref ("DW_AT_type", node);
    if (type_id == NULL) {
	type = strdup ("void %s");
    } else {
	xmlAttrPtr list = xmlGetID (xml_doc, type_id);
	assert (list != NULL);
	type = type_to_string (list->parent);
    }
    xmlFree (type_id);
    return type;
}

static char *
get_params (xmlNodePtr node)
{
    char *attr, *xpath;
    char buf [BUFLEN] = {'\0'};
    xmlNodeSetPtr result;
    int  i, n;
    void *free_p;

    assert (node->type == XML_ELEMENT_NODE);
    assert (!strcmp (node->name, "tag"));
    attr = xmlGetProp (node, "name");
    assert (!strcmp (attr, "DW_TAG_subroutine_type")
	    || !strcmp (attr, "DW_TAG_subprogram"));
    xmlFree (attr);

    xpath = "tag[@name='DW_TAG_unspecified_parameters'"
	    " or @name='DW_TAG_formal_parameter']";
    result = eval_xpath (xpath, node, &free_p, &n);

    for (i = 0; i < n; i++) {
	xmlNodePtr child = result->nodeTab [i];
	char *attr = xmlGetProp (child, "name");
	if (!strcmp (attr, "DW_TAG_unspecified_parameters")) {
	    strncat (buf, (i == 0 ? "" : ", ..."), sizeof (buf));
	} else if (!strcmp (attr, "DW_TAG_formal_parameter")) {
	    char buf2 [BUFLEN];
	    char *type = get_type (child);
	    char *name = get_attr_value ("DW_AT_name", child);
	    snprintf (buf2, sizeof (buf2), type, (name == NULL ? "" : name));
	    if (i != 0) strncat (buf, ", ", sizeof (buf));
	    strncat (buf, buf2, sizeof (buf));
	    free (type);
	    xmlFree (name);
	} else {
	    assert (0); exit (1);
	}
	xmlFree (attr);
    }

    free (free_p);

    if (n == 0)
	snprintf (buf, sizeof (buf), "void");
    return strdup (buf);
}

static char *
get_subscript        (xmlNodePtr node)
{
    /* Note: A DW_AT_upper_bound's value means 2 possibilities:
       (1) the maximum index value (e.g., 9 for 'int a[10]').
       (2) the pointer to the (artificial) DW_TAG_variable that
           holds an expression value for the array size.
       Problem here is that artificial DW_TAG_variable does not
       have a name, just a location.                             */
    int i;
    char buf [BUFLEN];
    char *ret, *form = NULL;
    char *abbrev_num = get_tag ("DW_TAG_subrange_type",
				"@abbrev_number", node);
    if (abbrev_num == NULL) {
	/* we cannot found the corresponding abbrev.
	   this might be a bug, but now just ignore it. */
    	return strdup ("");
    }

    for (i = 0; abbrev_table [i].number != -1; i++) {
	if (abbrev_table [i].number == atoi (abbrev_num)) {
	    form = abbrev_table [i].upper_bound_form;
	    break;
	}
    }

    xmlFree (abbrev_num);

    if (form == NULL) {
	ret = strdup ("");    /* no subscript information */
    } else if (!strncmp (form, "DW_FORM_ref", 11)) {
	ret = "<em>exp</em>";
    } else if (!strncmp (form, "DW_FORM_data", 12)) {
	xmlNodePtr child, child2;
	char *upper_bound = NULL;
	char *attr;

	for (child = node->children; child != NULL; child = child->next) {
	    if (child->type != XML_ELEMENT_NODE) continue;
	    if (strcmp (child->name, "tag"))     continue;
	    attr = xmlGetProp (child, "name");
	    if (strcmp (attr, "DW_TAG_subrange_type")) {
		xmlFree (attr);
		continue;
	    }
	    xmlFree (attr);

	    for (child2 = child->children; child2 != NULL;
		 child2 = child2->next) {
		if (child2->type != XML_ELEMENT_NODE)   continue;
		if (strcmp (child2->name, "attribute")) continue;
		attr = xmlGetProp (child2, "name");
		if (strcmp (attr, "DW_AT_upper_bound")) {
		    xmlFree (attr);
		    continue;
		}
		xmlFree (attr);
		upper_bound = xmlGetProp (child2, "value");
		break;
	    }
	}
	assert (upper_bound != NULL);
	snprintf (buf, sizeof (buf), "%d", atoi (upper_bound) + 1);
	xmlFree (upper_bound);
	ret = buf;
    } else {
	assert (0); exit (1);
    }
    return strdup (ret);
}


static int
get_comp_unit  (xmlNodePtr node)
{
    int i, ret;
    char *offset = xmlGetProp (node, "offset");
    unsigned long offset2;

    assert (offset != NULL);
    ret = sscanf (offset, "info:%ld", &offset2);
    assert (ret == 1);

    for (i = 0; i < n_comp_unit; i++) {
	if (offset2 < comp_unit [i].info_offset2)
	    break;
    }
    return i - 1;
}

static char *
get_filename_by_char (int n_comp, char *file_num)
{
    if (file_num == NULL) {
	return "??";
    } else {
	return get_filename_by_int (n_comp, atoi (file_num));
    }
}

static char *
get_filename_by_node (xmlNodePtr node, char *file_num)
{
    if (file_num == NULL) {
	return "??";
    } else {
	int n_comp = get_comp_unit (node);
	return get_filename_by_int (n_comp, atoi (file_num));
    }
}

static char *
get_filename_by_int  (int n_comp, int file_num)
{
    assert (0 <= n_comp && n_comp < n_comp_unit);
    assert (0 <= file_num && file_num <= comp_unit [n_comp].n_file);
    if (file_num > 0) {
	return comp_unit [n_comp].file [file_num].filename;
    } else {
	return "??";
    }
}

static char *
get_pathname_by_char (int n_comp, char *file_num)
{
    if (file_num == NULL) {
	return "??";
    } else {
	return get_pathname_by_int (n_comp, atoi (file_num));
    }
}

static char *
get_pathname_by_node (xmlNodePtr node, char *file_num)
{
    if (file_num == NULL) {
	return "??";
    } else {
	int n_comp = get_comp_unit (node);
	return get_pathname_by_int (n_comp, atoi (file_num));
    }
}

static char *
get_pathname_by_int  (int n_comp, int file_num)
{
    assert (0 <= n_comp && n_comp < n_comp_unit);
    assert (0 <= file_num && file_num <= comp_unit [n_comp].n_file);
    if (file_num > 0) {
	return comp_unit [n_comp].file [file_num].pathname;
    } else {
	return "??";
    }
}

static char *
get_line             (char *line)
{
    return (line == NULL ? "??" : line);
}

static char *
get_name             (char *name)
{
    return (name == NULL ? "??" : name);
}

static char *
type_to_string           (xmlNodePtr node)
{
    char *attr, *ret;
    assert (node->type == XML_ELEMENT_NODE);
    assert (!strcmp (node->name, "tag"));

    attr = xmlGetProp (node, "name");

    if (!strcmp (attr, "DW_TAG_array_type")) {
	ret = array_type_to_string (node);
    } else if (!strcmp (attr, "DW_TAG_base_type")) {
	ret = base_type_to_string (node);
    } else if (!strcmp (attr, "DW_TAG_pointer_type")) {
	ret = ptr_type_to_string (node);
    } else if (!strcmp (attr, "DW_TAG_subroutine_type")) {
	ret = func_type_to_string (node);
    } else if (!strcmp (attr, "DW_TAG_const_type")) {
	ret = const_to_string (node);
    } else if (!strcmp (attr, "DW_TAG_volatile_type")) {
	ret = volatile_to_string (node);
    } else if (!strcmp (attr, "DW_TAG_structure_type")) {
	ret = struct_to_string (node);
    } else if (!strcmp (attr, "DW_TAG_union_type")) {
	ret = union_to_string (node);
    } else if (!strcmp (attr, "DW_TAG_typedef")) {
	ret = typedef_to_string (node);
    } else if (!strcmp (attr, "DW_TAG_enumeration_type")) {
	ret = enum_to_string (node);
#if 0
    } else if (!strcmp (attr, "DW_TAG_member")) {
    } else if (!strcmp (attr, "DW_TAG_subrange_type")) {
#endif
    } else {
	assert (0); exit (1);
    }
    xmlFree (attr);
    return ret;
}
static char *
base_type_to_string    (xmlNodePtr node)
{
    char buf [BUFLEN];
    char *type = get_attr_value ("DW_AT_name", node);
    if (type == NULL) {
	snprintf (buf, sizeof (buf), "%s %%s", "void");
    } else {
	snprintf (buf, sizeof (buf), "%s %%s", type);
	xmlFree (type);
    }
    return strdup (buf);
}

static char *
array_type_to_string   (xmlNodePtr node)
{
    char buf [BUFLEN], buf2 [BUFLEN];
#if 0
    char *id        = xmlGetProp (node, "offset");
#endif
    char *type      = get_type (node);
    char *subscript = get_subscript (node);
    snprintf (buf2, sizeof (buf2), "%%s[%s]", subscript);
    snprintf (buf, sizeof (buf), type, buf2);
    free (subscript);
    free (type);
    return strdup (buf);
}

static char *
ptr_type_to_string   (xmlNodePtr node)
{
    char buf [BUFLEN];
    char *type = get_type (node);
    snprintf (buf, sizeof (buf), type, "(*%s)");
    free (type);
    return strdup (buf);
}

static char *
func_type_to_string  (xmlNodePtr node)
{
    char buf [BUFLEN], buf2 [BUFLEN];
    char *ret_type = get_type (node);
    char *params   = get_params (node);
    snprintf (buf2, sizeof (buf2), "(%%s)(%s)", params);
    snprintf (buf, sizeof (buf), ret_type, buf2);
    free (ret_type);
    free (params);
    return strdup (buf);
}

static char *
volatile_to_string   (xmlNodePtr node)
{
    char buf [BUFLEN];
    char *type = get_type (node);
    snprintf (buf, sizeof (buf), type, "volatile %s");
    free (type);
    return strdup (buf);
}

static char *
const_to_string      (xmlNodePtr node)
{
    char buf [BUFLEN];
    char *type = get_type (node);
    snprintf (buf, sizeof (buf), type, "const %s");
    free (type);
    return strdup (buf);
}

static char *
struct_union_to_string  (xmlNodePtr node, char *which)
{
    char buf [BUFLEN];
    char *id   = xmlGetProp (node, "offset");
    char *name = get_attr_value ("DW_AT_name", node);
    if (name == NULL) {
	char *members = members_to_string (node);
	snprintf (buf, sizeof (buf), "%s %s %%s", which, members);
    } else {
	snprintf (buf, sizeof (buf), "%s %s %%s", which, href_info (id, name));
	xmlFree (name);
    }
    xmlFree (id);
    return strdup (buf);
}

static char *
struct_to_string      (xmlNodePtr node)
{
    return struct_union_to_string (node, "struct");
}

static char *
union_to_string      (xmlNodePtr node)
{
    return struct_union_to_string (node, "union");
}

static char *
members_to_string    (xmlNodePtr node)
{
    char buf [BUFLEN] = {'\0'}, buf2 [BUFLEN], buf3 [BUFLEN];
    xmlNodeSetPtr result;
    int  i, n;
    char *xpath = "tag[@name='DW_TAG_member']";
    void *free_p;

    result = eval_xpath (xpath, node, &free_p, &n);

    strncat (buf, "{ ", sizeof (buf));

    for (i = 0; i < n; i++) {
	xmlNodePtr child = result->nodeTab [i];
	char *id       = xmlGetProp (child, "offset");
	char *name     = get_attr_value ("DW_AT_name", child);
	char *file     = get_attr_value ("DW_AT_decl_file", child);
	char *filename = get_filename_by_node (child, file);
	char *pathname = get_pathname_by_node (child, file);
#if 0
	char *line     = get_line (get_attr_value ("DW_AT_decl_line", child));
#endif
	char *type     = get_type (child);

	snprintf (buf2, sizeof (buf2), "%s", href_info (id, get_name (name)));
	snprintf (buf3, sizeof (buf3), type, buf2);
	strncat (buf, buf3, sizeof (buf));
	strncat (buf, "; ", sizeof (buf));
	xmlFree (id);
	xmlFree (name);
	xmlFree (file);
	free (type);
    }
    free (free_p);

    strncat (buf, "}", sizeof (buf));
    return strdup (buf);
}

static char *
typedef_to_string    (xmlNodePtr node)
{
    char buf [BUFLEN];
    char *id   = xmlGetProp (node, "offset");
    char *name = get_attr_value ("DW_AT_name", node);
    snprintf (buf, sizeof (buf), "%s %%s", href_info (id, get_name (name)));
    xmlFree (id);
    xmlFree (name);
    return strdup (buf);
}

static char *
enum_to_string       (xmlNodePtr node)
{
    char buf [BUFLEN];
    char *id      = xmlGetProp (node, "offset");
    char *name = get_attr_value ("DW_AT_name", node);
    if (name == NULL) {
	char * enum_constants = enum_constants_to_string (node);
	snprintf (buf, sizeof (buf), "enum %s %%s", enum_constants);
    } else {
	snprintf (buf, sizeof (buf), "enum %s %%s", href_info (id,  name));
    }
    xmlFree (id);
    xmlFree (name);
    return strdup (buf);
}

static char *
enum_constants_to_string (xmlNodePtr node)
{
    char buf [BUFLEN] = {'\0'}, buf2 [BUFLEN];
    xmlNodeSetPtr result;
    int  i, n;
    char *xpath = "tag[@name='DW_TAG_enumerator']";
    void *free_p;

    result = eval_xpath (xpath, node, &free_p, &n);
    strncat (buf, "{ ", sizeof (buf));

    for (i = 0; i < n; i++) {
	xmlNodePtr child = result->nodeTab [i];
	char *id    = xmlGetProp (child, "offset");
	char *name  = get_attr_value ("DW_AT_name", child);
	char *value = get_attr_value ("DW_AT_const_value", child);

	snprintf (buf2, sizeof (buf2),
		  "%s = %s, ", href_info (id, get_name (name)), value);
	strncat (buf, buf2, sizeof (buf));
	xmlFree (id);
	xmlFree (name);
	xmlFree (value);
    }

    free (free_p);
    strncat (buf, "}", sizeof (buf));
    return strdup (buf);
}

static void
output_indent         (FILE *fp)
{
    fprintf (fp, "%*s", indent_depth*4, "");
}

static void
visit_all_children (xmlNodePtr node, FILE *fp)
{
    xmlNodePtr child;
    for (child = node->children; child != NULL; child = child->next) {
	if (child->type != XML_ELEMENT_NODE || strcmp (child->name, "tag"))
	    continue;
	output_debug_info (child, fp);
    }
}

static void
output_debug_info_top (void)
{
    FILE *fp;
    xmlNodeSetPtr result;
    xmlNodePtr    node, child;
    void *free_p;

    if (args.verbose) printf ("processing .debug_info\n");
    result = eval_xpath ("//section[@name='.debug_info']",
			 xml_root, &free_p, NULL);
    assert (xmlXPathNodeSetGetLength (result) == 1);
    node = result->nodeTab [0];
    assert (node->type == XML_ELEMENT_NODE);
    assert (!strcmp (node->name, "section"));

    fp = fopen (WORKING_DIR XREF_INFO, "w");
    if (fp == NULL) {
	perror ("fopen (info)"); exit (1);
    }
    visit_all_children (node, fp);
    free (free_p);
    fclose (fp);
}

static void
output_debug_info (xmlNodePtr node, FILE *fp)
{
    char *attr;
    assert (node->type == XML_ELEMENT_NODE);
    assert (!strcmp (node->name, "tag"));
    attr = xmlGetProp (node, "name");
    if (args.verbose) printf ("\tprocessing %s\n", attr);

    if (!strcmp (attr, "DW_TAG_compile_unit")) {
	output_compile_unit (node, fp);
    } else if (!strcmp (attr, "DW_TAG_subprogram")) {
	output_subprogram (node, fp);
    } else if (!strcmp (attr, "DW_TAG_lexical_block")) {
	output_lexical_block (node, fp);
    } else if (!strcmp (attr, "DW_TAG_variable")) {
	output_variable (node, fp);
    } else if (!strcmp (attr, "DW_TAG_label")) {
	output_label (node, fp);
    } else if (!strcmp (attr, "DW_TAG_typedef")) {
	output_typedef (node, fp);
    } else if (!strcmp (attr, "DW_TAG_structure_type")) {
	output_struct (node, fp);
    } else if (!strcmp (attr, "DW_TAG_union_type")) {
	output_union (node, fp);
    } else if (!strcmp (attr, "DW_TAG_enumeration_type")) {
	output_enum (node, fp);
#if 0
    } else if (!strcmp (attr, "DW_TAG_array_type")
	       || !strcmp (attr, "DW_TAG_base_type")
	       || !strcmp (attr, "DW_TAG_const_type")
	       || !strcmp (attr, "DW_TAG_pointer_type")
	       || !strcmp (attr, "DW_TAG_subroutine_type")
	       || !strcmp (attr, "DW_TAG_volatile_type")
	       || !strcmp (attr, "DW_TAG_member")
	       || !strcmp (attr, "DW_TAG_subrange_type")
	       || !strcmp (attr, "DW_TAG_constant")
	       || !strcmp (attr, "DW_TAG_enumerator")
	       || !strcmp (attr, "DW_TAG_formal_parameter")
	       || !strcmp (attr, "DW_TAG_unspecified_parameters")) {
	visit_all_children (node, fp);
#endif
    } else {
	/*
	  just ignore for the following tags, not appearing in ANSI C:
	      DW_TAG_access_declaration DW_TAG_catch_block
	      DW_TAG_class_type DW_TAG_common_block
	      DW_TAG_common_inclusion DW_TAG_entry_point
	      DW_TAG_file_type DW_TAG_friend
	      DW_TAG_imported_declaration DW_TAG_inheritance
	      DW_TAG_inlined_subroutine DW_TAG_module
	      DW_TAG_namelist DW_TAG_namelist_item
	      DW_TAG_packed_type DW_TAG_ptr_to_member_type
	      DW_TAG_reference_type DW_TAG_set_type
	      DW_TAG_string_type
	      DW_TAG_template_type_param DW_TAG_template_value_param
	      DW_TAG_thrown_type DW_TAG_try_block
	      DW_TAG_variant DW_TAG_variant_part DW_TAG_with_stmt
	      and other vender-specific extentions.
	*/
    }
    xmlFree (attr);
}

static void
output_compile_unit (xmlNodePtr node, FILE *fp)
{
    char *id       = xmlGetProp (node, "offset");
    char *filename = get_attr_value ("DW_AT_name", node);
    char *comp_dir = get_attr_value ("DW_AT_comp_dir", node);
    char *high_pc  = get_attr_value ("DW_AT_high_pc", node);
    char *low_pc   = get_attr_value ("DW_AT_low_pc", node);
    char *producer = get_attr_value ("DW_AT_producer", node);

    fprintf (fp,
	     "</pre><hr/>\n"
	     "<h1>%s --- %s, 0x%x-0x%x, %s</h1><a name=\"%s\"/>\n"
	     "<hr/><pre>\n",
	     href_src1 (filename, "top", filename), comp_dir,
	     atoi (low_pc), atoi (high_pc), producer, id);
    visit_all_children (node, fp);

    xmlFree (id);
    xmlFree (filename);
    xmlFree (comp_dir);
    xmlFree (high_pc);
    xmlFree (low_pc);
    xmlFree (producer);
}

static void
output_subprogram     (xmlNodePtr node, FILE *fp)
{
    char buf [BUFLEN], buf2 [BUFLEN];
    char *id       = xmlGetProp (node, "offset");
    char *funcname = get_attr_value ("DW_AT_name", node);
    char *file     = get_attr_value ("DW_AT_decl_file", node);
    char *filename = get_filename_by_node (node, file);
    char *pathname = get_pathname_by_node (node, file);
    char *line     = get_attr_value ("DW_AT_decl_line", node);
    char *line2    = get_line (line);
    char *high_pc  = get_attr_value ("DW_AT_high_pc", node);
    char *low_pc   = get_attr_value ("DW_AT_low_pc", node);
    char *external = get_attr_value ("DW_AT_external", node);
    char *declaration = get_attr_value ("DW_AT_declaration", node);
    char *abst_origin = get_attr_value ("DW_AT_abstract_origin", node);
    char *inlined  = get_attr_desc ("DW_AT_inline", node);
    char *ret_type = get_type (node);
    char *params   = get_params (node);

    int  is_extern = (declaration != NULL) && strcmp (declaration, "0");

    /* skip this entry because this is a concrete code for inlined func */
    if (abst_origin != NULL)
	goto output_subprogram_exit;

    snprintf (buf2, sizeof (buf2), "%%s (%s)", params);
    snprintf (buf, sizeof (buf), ret_type, buf2);
    free (ret_type);
    free (params);

    snprintf (buf2, sizeof (buf2),
	      "%s<a name=\"%s\"/>",
	      href_src1 (pathname, line2, get_name (funcname)), id);

    if (inlined == NULL)
	inlined = "";

    output_indent (fp);
    if (low_pc != NULL && high_pc != NULL) {
	fprintf (fp, COMM_BEGIN "0x%x-0x%x, %s %s" COMM_END "\n",
		 atoi (low_pc), atoi (high_pc), 
		 href_src2 (pathname, line2, filename), inlined);
    } else {
	fprintf (fp, COMM_BEGIN "%s %s" COMM_END "\n",
		 href_src2 (pathname, line2, filename), inlined);
    }

    output_indent (fp);
    if (external == NULL || strcmp (external, "1"))
	fprintf (fp, "static ");
    if (is_extern)
	fprintf (fp, "extern ");
    fprintf (fp, buf, buf2);
    
    if (is_extern) {
	fprintf (fp, ";\n");
    } else {
	fprintf (fp, "\n{\n");
	indent_depth++;
	visit_all_children (node, fp);
	indent_depth--;
	fprintf (fp, "}\n", buf);
    }

output_subprogram_exit:
    xmlFree (id);
    xmlFree (funcname);
    xmlFree (file);
    xmlFree (line);
    xmlFree (high_pc);
    xmlFree (low_pc);
    xmlFree (external);
    xmlFree (declaration);
    xmlFree (abst_origin);
    xmlFree  (inlined);
}

static void
output_lexical_block  (xmlNodePtr node, FILE *fp)
{
    char *id       = xmlGetProp (node, "offset");
    char *high_pc  = get_attr_value ("DW_AT_high_pc", node);
    char *low_pc   = get_attr_value ("DW_AT_low_pc", node);
    output_indent (fp);
    fprintf (fp, 
	     "{ <a name=\"%s\"/><font color=\"red\">/* 0x%x-0x%x */</font>\n",
	     id,
	     (low_pc == NULL ? 0 : atoi (low_pc)),
	     (low_pc == NULL ? 0 : atoi (high_pc)));
    indent_depth++;
    visit_all_children (node, fp);
    indent_depth--;
    output_indent (fp);
    fprintf (fp, "}\n");

    xmlFree (id);
    xmlFree (high_pc);
    xmlFree (low_pc);
}

static void
output_variable       (xmlNodePtr node, FILE *fp)
{
    char buf [BUFLEN], loc [BUFLEN];
    int  offset, reg, is_auto = 0;
    unsigned long addr;
    char *id       = xmlGetProp (node, "offset");
    char *varname  = get_attr_value ("DW_AT_name", node);
    char *file     = get_attr_value ("DW_AT_decl_file", node);
    char *filename = get_filename_by_node (node, file);
    char *pathname = get_pathname_by_node (node, file);
    char *line     = get_attr_value ("DW_AT_decl_line", node);
    char *line2    = get_line (line);
    char *external = get_attr_value ("DW_AT_external", node);
    char *type     = get_type (node);
    char *location = get_attr_desc ("DW_AT_location", node);
    char *declaration = get_attr_value ("DW_AT_declaration", node);

    snprintf (buf, sizeof (buf), "%s%s",
	      href_src1 (pathname, line2, get_name (varname)), a_name (id));

    if (location == NULL) {
	snprintf (loc, sizeof (loc), "??");
    } else if (sscanf (location, "DW_OP_fbreg: %d", &offset) == 1) {
	/* allocated on stack */
	is_auto = 1;
	snprintf (loc, sizeof (loc), "%%fp+(%d)", offset);
    } else if (sscanf (location, "DW_OP_addr: %u", &addr) == 1) {
	/* allocated on data */
	snprintf (loc, sizeof (loc), "0x%x", addr);
    } else if (sscanf (location, "DW_OP_reg%d", &reg) == 1) {
	/* allocated on registers */
	is_auto = 1;
	snprintf (loc, sizeof (loc), "%%r%d", reg);
    } else {
	/* unknown description, so print it as it is. */
	snprintf (loc, sizeof (loc), "%s", location);
#if 0
	fprintf (stderr, "unknown description (%s)\n", location);
	assert (0);
#endif
    }

    output_indent (fp);
    if (!is_auto && (external == NULL || strcmp (external, "1")))
	fprintf (fp, "static ");
    if (declaration != NULL && strcmp (declaration, "0"))
	fprintf (fp, "extern ");
    fprintf (fp, type, buf);
    free (type);
    fprintf (fp, "; " COMM_BEGIN "%s, %s" COMM_END,
	     loc, href_src2 (pathname, line2, filename));
    fprintf (fp, "\n");

    xmlFree (id);
    xmlFree (varname);
    xmlFree (file);
    xmlFree (line);
    xmlFree (external);
    xmlFree (location);
    xmlFree (declaration);
}

static void
output_label (xmlNodePtr node, FILE *fp)
{
    char *id        = xmlGetProp (node, "offset");
    char *labelname = get_attr_value ("DW_AT_name", node);
    char *file      = get_attr_value ("DW_AT_decl_file", node);
    char *filename  = get_filename_by_node (node, file);
    char *pathname  = get_pathname_by_node (node, file);
    char *line      = get_attr_value ("DW_AT_decl_line", node);
    char *low_pc    = get_attr_value ("DW_AT_low_pc", node);

    output_indent (fp);
    fprintf (fp, "%s%s:",
	     href_src1 (pathname, get_line (line), get_name (labelname)),
	     a_name (id));
    fprintf (fp, " " COMM_BEGIN "0x%x, %s" COMM_END,
	     atoi (low_pc),
	     href_src2 (pathname, get_line (line), filename));
    fprintf (fp, "\n");

    xmlFree (id);
    xmlFree (labelname);
    xmlFree (file);
    xmlFree (line);
    xmlFree (low_pc);
}

static void
output_typedef  (xmlNodePtr node, FILE *fp)
{
    char buf [BUFLEN], loc [BUFLEN];
    char *id       = xmlGetProp (node, "offset");
    char *typename = get_attr_value ("DW_AT_name", node);
    char *file     = get_attr_value ("DW_AT_decl_file", node);
    char *filename = get_filename_by_node (node, file);
    char *pathname = get_pathname_by_node (node, file);
    char *line     = get_attr_value ("DW_AT_decl_line", node);
    char *type     = get_type (node);

    snprintf (buf, sizeof (buf), "%s%s",
	      href_src1 (pathname, get_line (line), get_name (typename)),
	      a_name (id));
    output_indent (fp);
    fprintf (fp, "typedef ");
    fprintf (fp, type, buf);
    free (type);
    fprintf (fp, "; ");
    fprintf (fp, " " COMM_BEGIN  "%s" COMM_END,
	     href_src2 (pathname, get_line (line), filename));
    fprintf (fp, "\n");

    xmlFree (id);
    xmlFree (typename);
    xmlFree (file);
    xmlFree (line);
}

static void
output_enum    (xmlNodePtr node, FILE *fp)
{
    char *id       = xmlGetProp (node, "offset");
    char *enumname = get_attr_value ("DW_AT_name", node);
    char *file     = get_attr_value ("DW_AT_decl_file", node);
    char *filename = get_filename_by_node (node, file);
    char *pathname = get_pathname_by_node (node, file);
    char *line     = get_attr_value ("DW_AT_decl_line", node);
    char *line2    = get_line (line);

    output_indent (fp);
    fprintf (fp, "enum %s%s " COMM_BEGIN "%s" COMM_END " ",
	     (enumname == NULL ? "" : href_src1 (pathname, line2, enumname)),
	     a_name (id), href_src2 (pathname, line2, filename));
    output_enum_constants (node, fp, pathname, line2);
    fprintf (fp, ";\n");

    xmlFree (id);
    xmlFree (enumname);
    xmlFree (file);
    xmlFree (line);
}

static void
output_enum_constants (xmlNodePtr node, FILE *fp, char *pathname, char *line)
{
    xmlNodeSetPtr result;
    int  i, n;
    char *xpath = "tag[@name='DW_TAG_enumerator']";
    void *free_p;

    result = eval_xpath (xpath, node, &free_p, &n);

    if (n == 0) return;

    output_indent (fp);
    fprintf (fp, "{\n");
    indent_depth++;

    for (i = 0; i < n; i++) {
	xmlNodePtr child = result->nodeTab [i];
	char *id    = xmlGetProp (child, "offset");
	char *name  = get_attr_value ("DW_AT_name", child);
	char *value = get_attr_value ("DW_AT_const_value", child);

	output_indent (fp);
	fprintf (fp, "%s = %s,%s\n", 
		 href_src1 (pathname, line, name), value, a_name (id));
	xmlFree (id);
	xmlFree (name);
	xmlFree (value);
    }
    free (free_p);
    indent_depth--;
    output_indent (fp);
    fprintf (fp, "}");
}

static void
output_struct_union (xmlNodePtr node, FILE *fp, char *which)
{
    char *id       = xmlGetProp (node, "offset");
    char *tagname  = get_attr_value ("DW_AT_name", node);
    char *file     = get_attr_value ("DW_AT_decl_file", node);
    char *filename = get_filename_by_node (node, file);
    char *pathname = get_pathname_by_node (node, file);
    char *line     = get_attr_value ("DW_AT_decl_line", node);
    char *line2    = get_line (line);
    char *size     = get_attr_value ("DW_AT_byte_size", node);

    output_indent (fp);
    fprintf (fp, "%s %s%s " COMM_BEGIN "%s" COMM_END " ", which, 
	     (tagname == NULL ? "" : href_src1 (pathname, line2, tagname)),
	     a_name (id), href_src2 (pathname, line2, filename));
    output_members (node, fp);
    fprintf (fp, ";\n");

    xmlFree (id);
    xmlFree (tagname);
    xmlFree (file);
    xmlFree (line);
    xmlFree (size);
}

static void
output_struct   (xmlNodePtr node, FILE *fp)
{
    output_struct_union (node, fp, "struct");
}

static void
output_union    (xmlNodePtr node, FILE *fp)
{
    output_struct_union (node, fp, "union");
}

static void
output_members  (xmlNodePtr node, FILE *fp)
{
    xmlNodeSetPtr result;
    int  i, n;
    char *xpath = "tag[@name='DW_TAG_member']";
    void *free_p;

    result = eval_xpath (xpath, node, &free_p, &n);

    if (n == 0) return;

    output_indent (fp);
    fprintf (fp, "{\n");
    indent_depth++;

    for (i = 0; i < n; i++) {
	char buf [BUFLEN];
	xmlNodePtr child = result->nodeTab [i];
	char *id       = xmlGetProp (child, "offset");
	char *name     = get_attr_value ("DW_AT_name", child);
	char *value    = get_attr_value ("DW_AT_const_value", child);
	char *file     = get_attr_value ("DW_AT_decl_file", child);
	char *filename = get_filename_by_node (child, file);
	char *pathname = get_pathname_by_node (child, file);
	char *line     = get_attr_value ("DW_AT_decl_line", child);
	char *line2    = get_line (line);
	char *location = get_attr_desc ("DW_AT_data_member_location", child);
	char *type     = get_type (child);

	snprintf (buf, sizeof (buf), "%s%s",
		  href_src1 (pathname, line2, get_name (name)), a_name (id));
	output_indent (fp);
	fprintf (fp, type, buf);

	if (location == NULL) {
	    fprintf (fp, "; " COMM_BEGIN "%s" COMM_END "\n",
		     href_src2 (pathname, line2, filename));
	} else {
	    int  offset;
	    if (sscanf (location, "DW_OP_plus_uconst: %d", &offset) != 1) {
		fprintf (stderr, "unknown description (%s)\n", location);
		assert (0);
	    }
	    fprintf (fp, "; " COMM_BEGIN "%d, %s" COMM_END "\n",
		     offset, href_src2 (pathname, line2, filename));
	}

	free (type);
	xmlFree (id);
	xmlFree (name);
	xmlFree (value);
	xmlFree (file);
	xmlFree (line);
	xmlFree (location);
    }
    free (free_p);
    indent_depth--;
    output_indent (fp);
    fprintf (fp, "}");
}

static void
output_all_structs    (void)
{
    FILE *fp1, *fp2;
    xmlNodeSetPtr result;
    int n, i;
    void *free_p;
    char *work1 = WORKING_DIR XREF_STRUCTS1;
    char *work2 = WORKING_DIR XREF_STRUCTS2;

    if (args.verbose) printf ("processing all structures\n");
    fp1 = fopen (work1, "w");
    if (fp1 == NULL) { perror ("fopen(structs1)"); exit (1); }
    fp2 = fopen (work2, "w");
    if (fp2 == NULL) { perror ("fopen(structs2)"); exit (1); }

    result = eval_xpath ("//tag[@name='DW_TAG_structure_type']",
			 xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	xmlNodePtr  node = result->nodeTab [i];
	char *id       = xmlGetProp (node, "offset");
	char *tagname  = get_attr_value ("DW_AT_name", node);
	char *file     = get_attr_value ("DW_AT_decl_file", node);
	char *filename = get_filename_by_node (node, file);
	char *pathname = get_pathname_by_node (node, file);
	char *line     = get_attr_value ("DW_AT_decl_line", node);
	char *line2    = get_line (line);

	if (tagname == NULL) goto loop_continue;
	fprintf (fp1, "%s @ struct %s; " COMM_BEGIN "%s" COMM_END "\n",
		 tagname, href_info (id, tagname), 
		 href_src2 (pathname, line2, filename));

	if (file == NULL) goto loop_continue;
	fprintf (fp2, "%s %s @ " COMM_BEGIN "%s" COMM_END " struct %s;\n",
		 filename, line2,
		 href_src2 (pathname, line2, filename),
		 href_info (id, tagname));

    loop_continue:
	xmlFree (id);
	xmlFree (tagname);
	xmlFree (file);
	xmlFree (line);
    }
    free (free_p);
    fclose (fp1);
    fclose (fp2);
}

static void
output_all_unions     (void)
{
    FILE *fp1, *fp2;
    xmlNodeSetPtr result;
    int n, i;
    void *free_p;
    char *work1 = WORKING_DIR XREF_UNIONS1;
    char *work2 = WORKING_DIR XREF_UNIONS2;

    if (args.verbose) printf ("processing all unions\n");
    fp1 = fopen (work1, "w");
    if (fp1 == NULL) { perror ("fopen(unions1)"); exit (1); }
    fp2 = fopen (work2, "w");
    if (fp2 == NULL) { perror ("fopen(unions2)"); exit (1); }

    result = eval_xpath ("//tag[@name='DW_TAG_union_type']",
			 xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	xmlNodePtr  node = result->nodeTab [i];
	char *id       = xmlGetProp (node, "offset");
	char *tagname  = get_attr_value ("DW_AT_name", node);
	char *file     = get_attr_value ("DW_AT_decl_file", node);
	char *filename = get_filename_by_node (node, file);
	char *pathname = get_pathname_by_node (node, file);
	char *line     = get_attr_value ("DW_AT_decl_line", node);
	char *line2    = get_line (line2);

	if (tagname == NULL) goto loop_continue;
	fprintf (fp1, "%s @ union %s; " COMM_BEGIN "%s" COMM_END "\n",
		 tagname, href_info (id, tagname), 
		 href_src2 (pathname, line2, filename));
	if (file == NULL) goto loop_continue;
	fprintf (fp2, "%s %s @ " COMM_BEGIN "%s" COMM_END " union %s;\n",
		 filename, line2,
		 href_src2 (pathname, line2, filename),
		 href_info (id, tagname));

    loop_continue:
	xmlFree (id);
	xmlFree (tagname);
	xmlFree (file);
	xmlFree (line);
    }
    free (free_p);
    fclose (fp1);
    fclose (fp2);
}

static void
output_all_typedefs   (void)
{
    FILE *fp1, *fp2;
    xmlNodeSetPtr result;
    int n, i;
    void *free_p;
    char *work1 = WORKING_DIR XREF_TYPEDEFS1;
    char *work2 = WORKING_DIR XREF_TYPEDEFS2;

    if (args.verbose) printf ("processing all typedefs\n");
    fp1 = fopen (work1, "w");
    if (fp1 == NULL) { perror ("fopen(tdefs1)"); exit (1); }
    fp2 = fopen (work2, "w");
    if (fp2 == NULL) { perror ("fopen(tdefs2)"); exit (1); }

    result = eval_xpath ("//tag[@name='DW_TAG_typedef']",
			 xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	xmlNodePtr  node = result->nodeTab [i];
	char *id       = xmlGetProp (node, "offset");
	char *name     = get_attr_value ("DW_AT_name", node);
	char *file     = get_attr_value ("DW_AT_decl_file", node);
	char *filename = get_filename_by_node (node, file);
	char *pathname = get_pathname_by_node (node, file);
	char *line     = get_attr_value ("DW_AT_decl_line", node);
	char *line2    = get_line (line);

	if (name == NULL) goto loop_continue;
	fprintf (fp1, "%s @ typedef %s; " COMM_BEGIN "%s" COMM_END "\n",
		 name, href_info (id, name),
		 href_src2 (pathname, line2, filename));
	if (file == NULL) goto loop_continue;
	fprintf (fp2, "%s %s @ " COMM_BEGIN "%s" COMM_END " union %s;\n",
		 filename, line2,
		 href_src2 (pathname, line2, filename),
		 href_info (id, name));

    loop_continue:
	xmlFree (id);
	xmlFree (name);
	xmlFree (file);
	xmlFree (line);
    }
    free (free_p);
    fclose (fp1);
    fclose (fp2);
}

static void
output_all_enum_tags      (void)
{
    FILE *fp1, *fp2;
    xmlNodeSetPtr result;
    int n, i;
    void *free_p;
    char *work1 = WORKING_DIR XREF_ENUM_TAGS1;
    char *work2 = WORKING_DIR XREF_ENUM_TAGS2;

    if (args.verbose) printf ("processing all enum tags\n");
    fp1 = fopen (work1, "w");
    if (fp1 == NULL) { perror ("fopen(etags1)"); exit (1); }
    fp2 = fopen (work2, "w");
    if (fp2 == NULL) { perror ("fopen(etags2)"); exit (1); }

    result = eval_xpath ("//tag[@name='DW_TAG_enumeration_type']",
			 xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	xmlNodePtr  node = result->nodeTab [i];
	char *id       = xmlGetProp (node, "offset");
	char *name     = get_attr_value ("DW_AT_name", node);
	char *file     = get_attr_value ("DW_AT_decl_file", node);
	char *filename = get_filename_by_node (node, file);
	char *pathname = get_pathname_by_node (node, file);
	char *line     = get_attr_value ("DW_AT_decl_line", node);
	char *line2    = get_line (line);

	if (name == NULL) goto loop_continue;
	fprintf (fp1, "%s @ enum %s; " COMM_BEGIN "%s" COMM_END "\n",
		 name, href_info (id, name),
		 href_src2 (pathname, line2, filename));
	if (file == NULL) goto loop_continue;
	fprintf (fp2, "%s %s @ " COMM_BEGIN "%s" COMM_END " enum %s;\n",
		 filename, line2,
		 href_src2 (pathname, line2, filename),
		 href_info (id, name));

    loop_continue:
	xmlFree (id);
	xmlFree (name);
	xmlFree (file);
	xmlFree (line);
    }
    free (free_p);
    fclose (fp1);
    fclose (fp2);
}

static void
output_all_enum_consts      (void)
{
    FILE *fp1, *fp2;
    xmlNodeSetPtr result;
    int n, i;
    void *free_p;
    char *work1 = WORKING_DIR XREF_ENUM_CONSTS1;
    char *work2 = WORKING_DIR XREF_ENUM_CONSTS2;

    if (args.verbose) printf ("processing all enum constants\n");
    fp1 = fopen (work1, "w");
    if (fp1 == NULL) { perror ("fopen(econs1)"); exit (1); }
    fp2 = fopen (work2, "w");
    if (fp2 == NULL) { perror ("fopen(econs2)"); exit (1); }

    result = eval_xpath ("//tag[@name='DW_TAG_enumerator']",
			 xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	xmlNodePtr  node = result->nodeTab [i];
	char *id       = xmlGetProp (node, "offset");
	char *name     = get_attr_value ("DW_AT_name", node);
	char *file     = get_attr_value ("DW_AT_decl_file", node->parent);
	char *filename = get_filename_by_node (node, file);
	char *pathname = get_pathname_by_node (node, file);
	char *line     = get_attr_value ("DW_AT_decl_line", node->parent);
	char *line2    = get_line (line);

	if (name == NULL) goto loop_continue;
	fprintf (fp1, "%s @ %s; " COMM_BEGIN "%s" COMM_END "\n",
		 name, href_info (id, name),
		 href_src2 (pathname, line2, filename));
	if (file == NULL) goto loop_continue;
	fprintf (fp2, "%s %s @ " COMM_BEGIN "%s" COMM_END " %s;\n",
		 filename, line2,
		 href_src2 (pathname, line2, filename),
		 href_info (id, name));

    loop_continue:
	xmlFree (id);
	xmlFree (name);
	xmlFree (file);
	xmlFree (line);
    }
    free (free_p);
    fclose (fp1);
    fclose (fp2);
}

static void
output_all_labels     (void)
{
    FILE *fp1, *fp2;
    xmlNodeSetPtr result;
    int n, i;
    void *free_p;
    char *work1 = WORKING_DIR XREF_LABELS1;
    char *work2 = WORKING_DIR XREF_LABELS2;

    if (args.verbose) printf ("processing all labels\n");
    fp1 = fopen (work1, "w");
    if (fp1 == NULL) { perror ("fopen(labels1)"); exit (1); }
    fp2 = fopen (work2, "w");
    if (fp2 == NULL) { perror ("fopen(labels2)"); exit (1); }

    result = eval_xpath ("//tag[@name='DW_TAG_label']",
			 xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	xmlNodePtr  node = result->nodeTab [i];
	char *id       = xmlGetProp (node, "offset");
	char *name     = get_attr_value ("DW_AT_name", node);
	char *file     = get_attr_value ("DW_AT_decl_file", node);
	char *filename = get_filename_by_node (node, file);
	char *pathname = get_pathname_by_node (node, file);
	char *line     = get_attr_value ("DW_AT_decl_line", node);
	char *line2    = get_line (line);

	if (name == NULL) goto loop_continue;
	fprintf (fp1, "%s @ %s: " COMM_BEGIN "%s" COMM_END "\n",
		 name, href_info (id, name),
		 href_src2 (pathname, line2, filename));
	if (file == NULL) goto loop_continue;
	fprintf (fp2, "%s %s @ " COMM_BEGIN "%s" COMM_END " %s:\n",
		 filename, line2, href_src2 (pathname, line2, filename),
		 href_info (id, name));

    loop_continue:
	xmlFree (id);
	xmlFree (name);
	xmlFree (file);
	xmlFree (line);
    }
    free (free_p);
    fclose (fp1);
    fclose (fp2);
}

static void
output_all_functions  (void)
{
    FILE *fp1, *fp2;
    xmlNodeSetPtr result;
    int n, i;
    void *free_p;
    char *work1 = WORKING_DIR XREF_FUNCTIONS1;
    char *work2 = WORKING_DIR XREF_FUNCTIONS2;

    if (args.verbose) printf ("processing all functions\n");
    fp1 = fopen (work1, "w");
    if (fp1 == NULL) { perror ("fopen(funcs1)"); exit (1); }
    fp2 = fopen (work2, "w");
    if (fp2 == NULL) { perror ("fopen(funcs2)"); exit (1); }

    result = eval_xpath ("//tag[@name='DW_TAG_subprogram']",
			 xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	char buf [BUFLEN], buf2 [BUFLEN];
	xmlNodePtr  node = result->nodeTab [i];
	char *id       = xmlGetProp (node, "offset");
	char *name     = get_attr_value ("DW_AT_name", node);
	char *file     = get_attr_value ("DW_AT_decl_file", node);
	char *filename = get_filename_by_node (node, file);
	char *pathname = get_pathname_by_node (node, file);
	char *line     = get_attr_value ("DW_AT_decl_line", node);
	char *line2    = get_line (line);
	char *ret_type = get_type (node);
	char *params   = get_params (node);
	char *external = get_attr_value ("DW_AT_external", node);

	if (name == NULL) goto loop_continue;

	snprintf (buf2, sizeof (buf2), "%%s (%s)", params);
	snprintf (buf, sizeof (buf), ret_type, buf2);
	free (ret_type);
	free (params);

	fprintf (fp1, "%s @ ", name);
	if (external != NULL && !strcmp (external, "0"))
	    fprintf (fp1, "static ");
	snprintf (buf2, sizeof (buf2), "%s", href_info (id, name));
	fprintf (fp1, buf, buf2);
	fprintf (fp1, "; " COMM_BEGIN "%s" COMM_END "\n",
		 href_src2 (pathname, line2, filename));

	if (file == NULL) goto loop_continue;

	fprintf (fp2, "%s %s @ ", filename, pathname);
	fprintf (fp2, COMM_BEGIN "%s" COMM_END " ",
		 href_src2 (pathname, line2, filename));
	if (external != NULL && !strcmp (external, "0"))
	    fprintf (fp2, "static ");
	snprintf (buf2, sizeof (buf2), "%s", href_info (id, name));
	fprintf (fp2, buf, buf2);
	fprintf (fp2, ";\n");

    loop_continue:
	xmlFree (id);
	xmlFree (name);
	xmlFree (file);
	xmlFree (line);
	xmlFree (external);
    }
    free (free_p);
    fclose (fp1);
    fclose (fp2);
}

static void
output_all_variables  (void)
{
    FILE *fp1, *fp2;
    xmlNodeSetPtr result;
    int n, i;
    void *free_p;
    char *work1 = WORKING_DIR XREF_VARIABLES1;
    char *work2 = WORKING_DIR XREF_VARIABLES2;

    if (args.verbose) printf ("processing all variables\n");
    fp1 = fopen (work1, "w");
    if (fp1 == NULL) { perror ("fopen(vars1)"); exit (1); }
    fp2 = fopen (work2, "w");
    if (fp2 == NULL) { perror ("fopen(vars2)"); exit (1); }

    result = eval_xpath ("//tag[@name='DW_TAG_variable']",
			 xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	char buf [BUFLEN];
	xmlNodePtr  node = result->nodeTab [i];
	char *id       = xmlGetProp (node, "offset");
	char *name     = get_attr_value ("DW_AT_name", node);
	char *file     = get_attr_value ("DW_AT_decl_file", node);
	char *filename = get_filename_by_node (node, file);
	char *pathname = get_pathname_by_node (node, file);
	char *line     = get_attr_value ("DW_AT_decl_line", node);
	char *line2    = get_line (line);
	char *type = get_type (node);
	char *external = get_attr_value ("DW_AT_external", node);

	if (name == NULL) goto loop_continue;

	fprintf (fp1, "%s @ ", name);
	if (external != NULL && !strcmp (external, "0"))
	    fprintf (fp1, "static ");
	snprintf (buf, sizeof (buf), "%s", href_info (id, name));
	fprintf (fp1, type, buf);
	fprintf (fp1, "; " COMM_BEGIN "%s" COMM_END "\n",
		 href_src2 (pathname, line2, filename));

	if (file == NULL) goto loop_continue;

	fprintf (fp2, "%s %s @ ", filename, line2);
	fprintf (fp2, COMM_BEGIN "%s" COMM_END " ",
		 href_src2 (pathname, line2, filename));
	if (external != NULL && !strcmp (external, "0"))
	    fprintf (fp2, "static ");
	snprintf (buf, sizeof (buf), "%s", href_info (id, name));
	fprintf (fp2, type, buf);
	fprintf (fp2, ";\n");

    loop_continue:
	xmlFree (id);
	xmlFree (name);
	xmlFree (file);
	xmlFree (line);
	xmlFree (external);
	free (type);
    }
    free (free_p);
    fclose (fp1);
    fclose (fp2);
}

static void
output_all_files      (void)
{
    FILE *fp1;
    xmlNodeSetPtr result;
    int n, i;
    void *free_p;
    char *work1 = WORKING_DIR XREF_FILES1;

    if (args.verbose) printf ("processing all files\n");
    fp1 = fopen (work1, "w");
    if (fp1 == NULL) { perror ("fopen(files1)"); exit (1); }

    result = eval_xpath ("//tag[@name='DW_TAG_compile_unit']",
			 xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	xmlNodePtr  node = result->nodeTab [i];
	char *id       = xmlGetProp (node, "offset");
	char *name     = get_attr_value ("DW_AT_name", node);

	if (name == NULL) goto loop_continue;

	fprintf (fp1, "%s @ %s", name, href_info (id, name));
	fprintf (fp1, " " COMM_BEGIN "%s" COMM_END "\n",
		 href_src1 (name, "top", name));

    loop_continue:
	xmlFree (id);
	xmlFree (name);
    }
    free (free_p);
    fclose (fp1);
}

static char * 
get_macinfo_id (void)
{
    static char buf [BUFLEN];
    static int counter = 0;
    snprintf (buf, sizeof (buf), "MACINFOID%d", counter++);
    return buf;
}

static void
output_all_macros     (void)
{
    FILE *fp1, *fp2, *fp3, *fp4;
    xmlNodeSetPtr result;
    int n, i;
    int curr_comp_unit = -1;
    void *free_p;
    char *work1 = WORKING_DIR XREF_MACROS1;
    char *work2 = WORKING_DIR XREF_MACROS2;
    char *work3 = WORKING_DIR XREF_MACROS3;
    char *work4 = WORKING_DIR XREF_MACINFO;

    enum { STACK_SIZE = 1024 };
    int file_stack [STACK_SIZE], sp = 0;

    if (args.verbose) printf ("processing all macros\n");
    fp1 = fopen (work1, "w");
    if (fp1 == NULL) { perror ("fopen(macros1)"); exit (1); }
    fp2 = fopen (work2, "w");
    if (fp2 == NULL) { perror ("fopen(macros2)"); exit (1); }
    fp3 = fopen (work3, "w");
    if (fp3 == NULL) { perror ("fopen(macros3)"); exit (1); }
    fp4 = fopen (work4, "w");
    if (fp4 == NULL) { perror ("fopen(macros4)"); exit (1); }

    result = eval_xpath ("//section[@name='.debug_macinfo']/*",
			 xml_root, &free_p, &n);
    /* for font-lock sanity */

    for (i = 0; i < n; i++) {
	char buf [BUFLEN];
	xmlNodePtr  node = result->nodeTab [i];

	if (node->type != XML_ELEMENT_NODE) continue;

	if (!strcmp (node->name, "start_file")) {
	    char *line = xmlGetProp (node, "line");
	    char *file_num = xmlGetProp (node, "file_index");
	    char *filename, *pathname;
	    if (sp >= STACK_SIZE) {
		fprintf (stderr, "file stack overflow\n");
		exit (1);
	    }
	    if (sp == 0) {
		int j;
		char *offset = xmlGetProp (node, "offset");
		assert (offset != NULL);
		
		if (strcmp (line, "0")) {
		    if (line == NULL) line = "(null)";
		    printf ("line = %s, offset = %s\n", line, offset);
		}
		assert (!strcmp (line, "0"));

		for (j = 0; j < n_comp_unit; j++) {
		    if (!strcmp (offset, comp_unit [j].macinfo_offset)) {
			curr_comp_unit = j;
			break;
		    }
		}
		/* not found, so skip this comp_unit */
		if (j >= n_comp_unit) {
		    while (1) {
			char *line;
			i++;
			if (i >= n) break;
			node = result->nodeTab [i];
			if (strcmp (node->name, "start_file"))
			    continue;
			line = xmlGetProp (node, "line");
			if (strcmp (line, "0")) {
			    xmlFree (line);
			    continue;
			}
			xmlFree (line);
			break;
		    }
#if 0
		    xmlAttrPtr attr;
		    unsigned long offset2;
		    sscanf (offset, "macinfo:%ld", &offset2);
		    for (j = 0; j < n_comp_unit; j++) {
			if (offset2 > comp_unit [j].macinfo_offset2)
			    break;
		    }
		    attr = xmlGetID (xml_doc, comp_unit [i].macinfo_offset);
		    assert (attr != NULL);
		    node = attr->parent;
#endif
		    i--;
		    goto start_file_exit;
		}
	    }
	    assert (file_num != NULL);
	    file_stack [sp++] = atoi (file_num);
	    filename = get_filename_by_int (curr_comp_unit,
					    file_stack [sp - 1]);
	    pathname = get_pathname_by_int (curr_comp_unit,
					    file_stack [sp - 1]);

	    if (sp < 2) {
		fprintf (fp4, "<hr><h1>%s, %s</h1><hr>",
			 filename, pathname);
	    } else {
		char *id = get_macinfo_id ();
		char *filename2 = get_filename_by_int (curr_comp_unit,
						       file_stack [sp - 2]);
		char *pathname2 = get_pathname_by_int (curr_comp_unit,
						       file_stack [sp - 2]);
		fprintf (fp1,
			 "%s @ #include %s " COMM_BEGIN "%s" COMM_END "\n",
			 filename, href_macinfo (id, filename),
			 href_src2 (pathname2, line, filename2));
		fprintf (fp2,
			 "a%s @ #include %s " COMM_BEGIN "%s" COMM_END "\n",
			 filename, href_macinfo (id, filename),
			 href_src2 (pathname2, line, filename2));
		fprintf (fp3,
			 "%s %s @ " COMM_BEGIN "%s" COMM_END " #include %s\n",
			 filename, line,
			 href_src2 (pathname2, line, filename2),
			 href_macinfo (id, filename));
		fprintf (fp4,
			 "%s#include %s " COMM_BEGIN "%s" COMM_END "\n",
			 a_name (id), href_src1 (pathname2, line, filename),
			 href_src2 (pathname2, line, filename2));
	    }

	start_file_exit:
	    xmlFree (line);
	    xmlFree (file_num);
	} else if (!strcmp (node->name, "end_file")) {
	    char *filename = get_filename_by_int (curr_comp_unit,
						  file_stack [sp - 1]);
	    if (sp < 0) {
		fprintf (stderr, "file stack underflow\n");
		exit (1);
	    }
	    sp--;
	} else if (!strcmp (node->name, "define_func")
		   || !strcmp (node->name, "define_normal")) {
	    char *id       = get_macinfo_id ();
	    char *line     = xmlGetProp (node, "line");
	    char *line2    = get_line (line);
	    char *name     = xmlGetProp (node, "name");
	    char *filename = get_filename_by_int (curr_comp_unit,
						  file_stack [sp - 1]);
	    char *pathname = get_pathname_by_int (curr_comp_unit,
						  file_stack [sp - 1]);
	    char *body     = xmlGetProp (node, "body");
	    char *signature = xmlGetProp (node, "signature");

	    fprintf (fp1, "%s @ #define %s " COMM_BEGIN "%s" COMM_END "\n",
		     name, href_macinfo (id, signature),
		     href_src2 (pathname, line2, filename));
	    fprintf (fp2, "d%s @ #define %s " COMM_BEGIN "%s" COMM_END "\n",
		     name, href_macinfo (id, signature),
		     href_src2 (pathname, line2, filename));
	    fprintf (fp3, "%s %s @ " COMM_BEGIN "%s" COMM_END " #define %s\n",
		     filename, line2, href_src2 (pathname, line2, filename),
		     href_macinfo (id, signature));
	    fprintf (fp4, "%s#define %s %s " COMM_BEGIN "%s" COMM_END "\n",
		     a_name (id),
		     href_src1 (pathname, line2, signature),
		     body,
		     href_src2 (pathname, line2, filename));

	    xmlFree (line);
	    xmlFree (name);
	    xmlFree (body);
	    xmlFree (signature);
	} else if (!strcmp (node->name, "undef")) {
	    char *id       = get_macinfo_id ();
	    char *line     = xmlGetProp (node, "line");
	    char *line2    = get_line (line);
	    char *name     = xmlGetProp (node, "name");
	    char *filename = get_filename_by_int (curr_comp_unit,
						  file_stack [sp - 1]);
	    char *pathname = get_pathname_by_int (curr_comp_unit,
						  file_stack [sp - 1]);

	    fprintf (fp1, "%s @ #undef %s " COMM_BEGIN "%s" COMM_END "\n",
		     name, href_macinfo (id, name),
		     href_src2 (pathname, line2, filename));
	    fprintf (fp2, "u%s @ #undef %s " COMM_BEGIN "%s" COMM_END "\n",
		     name, href_macinfo (id, name),
		     href_src2 (pathname, line2, filename));
	    fprintf (fp3, "%s %s @ " COMM_BEGIN "%s" COMM_END " #undef %s\n",
		     filename, line2, href_src2 (pathname, line2, filename),
		     href_macinfo (id, name));
	    fprintf (fp4, "%s#undef %s " COMM_BEGIN "%s" COMM_END "\n",
		     a_name (id),
		     href_src1 (pathname, line2, name),
		     href_src2 (pathname, line2, filename));

	    xmlFree (line);
	    xmlFree (name);
	}
    }

    free (free_p);
    fclose (fp1);
    fclose (fp2);
    fclose (fp3);
    fclose (fp4);
}
/* ====================================================================== */
int
main (int argc, char *argv [])
{
    if (argc <= 1) {
	print_usage ();
	exit (1);
    }
    process_args (argc, argv);
    init ();

    output_debug_info_top   ();
    output_all_structs      ();
    output_all_unions       ();
    output_all_typedefs     ();
    output_all_enum_tags    ();
    output_all_enum_consts  ();
    output_all_labels       ();
    output_all_functions    ();
    output_all_variables    ();
    output_all_files        ();
    output_all_macros       ();

    final ();
    return 0;
}
/* ====================================================================== */
